/* 	=================================================================================

Demonstration of programming a microcontroller application using FSM style 

Fisrt design at: http://digsys.upc.es/ed//CSD/units/Ch4/U4_10/Heads_Tails.zip

Adding the LCD peripheral to show betting results. Two new states have been added 
("Write_Heads", "Write_Tails") to allow writing the LCD before going idle once the Button is released.

In the same way, other states can be added to show other messages before starting the 
square wave generation

================================================================================= */
  	
#include	<pic.h>		// This file also includes the "pic1687x.h" which contains all the 
						// microcontroller declarations 
#include <stdio.h>
#include "lcd.h"

/* 	=================================================================================
	Definitions 
   	================================================================================= */
// This structure allows to access to a given bit into an 8-bit register
#define 	PORTBIT(adr, bit) ( (unsigned)(&adr)*8 + (bit) )

/* This is the state enumeration. Check the state diagram posted on the web to figure out which 
is the meaning of every state */
#define 	Idle_Low 		1
#define 	Set_Value_High	2		
#define 	Wave_High 		3
#define 	Idle_High		4
#define 	Set_Value_Low	5		
#define 	Wave_Low 		6
#define 	Write_Heads		7
#define 	Write_Tails		8


// Other constants:
#define 	Timer_Count_High 15						
#define 	Timer_Count_Low  15

/* 	=================================================================================
	Function prototypes 
   	================================================================================= */
static void interrupt service_routine (void);		// This is the interrupt service routine 

void init_system (void);
char output_logic(void);
char state_logic(void);
char read_inputs (void);



/* 	=================================================================================
	Global variables  
  	================================================================================= */
static bit Out_Z @ PORTBIT(PORTB, 3); 	// LED 3 at the PICDEM2+ will generate the waveform
static bit Buzzer @ PORTBIT(PORTC, 2); 	// Buzzer RC2 --> when working at high frequency
										// the square wave generates audible sound


static bit pushbutton_L @ PORTBIT(PORTB, 0); // Pushbutton  RB0

static char present_state = Idle_Low;		// state variable 

static char Button;				

static bit Time_Out = 0;		// The flag to indicate that the Timer0 has finished

/* 	=================================================================================
	Main function 
  	================================================================================= */
void main(void){

	init_system ();
	Out_Z = 0;	
								/* loop forever */
	while(1) {
		output_logic();
		state_logic();
	}		

}	

/* 	=================================================================================
	Function definitions
  	================================================================================= */

//*****************************************************************************
//State_logic routine 
//*****************************************************************************
char state_logic(void){
	char error = 0;

	Button = read_inputs(); 

	switch (present_state) {
		case Idle_Low:				
			if (Button == 1) {				// if a button pressed is detected,
				present_state = Set_Value_High; 			
			}
			else{
				present_state = Idle_Low; 
			}
			break;
	
		case Idle_High:						
			if (Button == 1) {	
				present_state = Set_Value_Low; 			
			}
			else{
				present_state = Idle_High; 
			}
			break;		

		case Wave_Low:
		
			if (Button == 0) 
				present_state = Write_Heads; 						
			else {
				if (Time_Out == 0) 
					present_state = Wave_Low; 			
				else
					present_state = Set_Value_High;
			}			
			break;

		case Wave_High:			
			if (Button ==0) 
				present_state = Write_Tails; 						
			else {
				if (Time_Out == 0) 
					present_state = Wave_High; 			
				else
					present_state = Set_Value_Low;
			}
			break;	

		case Set_Value_High:			
				present_state = Wave_High; 	
			break;

		case Set_Value_Low:			
				present_state = Wave_Low; 	
			break;

		case Write_Heads:			
				present_state =Idle_Low; 	
			break;

		case Write_Tails:			
				present_state =Idle_High; 	
			break;

		default:
			error = 1;
	}	
	return (error);
}

//*****************************************************************************
//Output logic routine 
//*****************************************************************************
char output_logic(void){
unsigned char error = 0;

	switch (present_state) {	
		case Idle_Low:			
			Out_Z = 0; 
			break;

		case Idle_High:		
			Out_Z = 1; 
			break;

		case Write_Heads:
				lcd_clear();
				lcd_home();						// select line 1
				lcd_puts("Your bet is ...");
				lcd_home2();					// select line 2 to write the result 
				lcd_puts(" ... HEADS ...");	
			break;

		case Write_Tails:
				lcd_clear();
				lcd_home();						// select line 1
				lcd_puts("Your bet is ...");
				lcd_home2();					// select line 2 to write the result 
				lcd_puts(" ... TAILS ...");	
			break;

		case Wave_Low:			
			Out_Z = 0; 
			break;

		case Wave_High:			
			Out_Z = 1; 
			break;

		case Set_Value_High:					// Trigger the Timer0 and enable its interrupts 	
			Time_Out = 0;					
			TMR0 = (256 - Timer_Count_High);		
			T0IE = 1;						
			Out_Z = 1; 

			break;

		case Set_Value_Low:						// Trigger the Timer0 and enable its interrupts 
			Time_Out = 0;
			TMR0 = (256 - Timer_Count_Low);	
			T0IE = 1;			
			Out_Z = 0; 
			break;
								
		default:
			error = 1;
			break;		
	}
	Buzzer = Out_Z; 	// Just a copy of the LED out for the buzzer circuit at RC2

	return (error);
}

//*****************************************************************************
//Init operations  
//*****************************************************************************
void init_system (void){


 /* These are the PIC16F877A configuration bits translated to the XC8 Compiler.
    They can be generated automatically in the MPLAB X IDE, but normally,
    you don't have to change them.
    So, copy and paste them as they are to your projects */

//#pragma config FOSC = EXTRC   // Oscillator Selection bits (RC oscillator)
#pragma config FOSC = XT     	// External XTAL Oscillator

//#pragma config WDTE = ON      // Watchdog Timer Enable bit (WDT enabled)
#pragma config WDTE = OFF       // Watchdog Timer not in use (WDT disabled)
//#pragma config PWRTE = OFF    // Power-up Timer Enable bit (PWRT disabled)

#pragma config PWRTE = ON      	// Power-up Timer Enable bit (PWRT enabled)

#pragma config BOREN = ON       // Brown-out Reset Enable bit (BOR enabled)
#pragma config LVP = ON         // Low-Voltage (Single-Supply) In-Circuit Serial
				// Programming Enable bit (RB3/PGM pin has PGM
				// function; low-voltage programming enabled)

#pragma config CPD = OFF        // Data EEPROM Memory Code Protection bit
        			//(Data EEPROM code protection off)
#pragma config WRT = OFF        // Flash Program Memory Write Enable bits (Write
				// protection off; all program memory may be
				// written to by EECON control)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit
                		// (Code protection off)

// Inicialise starts here:
	PORTA 	= 0x00; // Reset all Flip-Flops at PORTA 

	ADCON1 	= 0x06;		/* All the PortA is digital */

	TRISA 	= 0x3F;		/* All inputs */

	PORTC 	= 0x00; 	// Reset all Flip-Flops at PORTC 
	TRISC 	= 0xFB;		/* RC2 is output */

	PORTB 	= 0x00; // Reset all Flip-Flops at PORTB 
	TRISB 	= 0x01; // PORTB are outputs except for RB0, which is the external INT

	INTEDG	= 0;	// INT on falling edge (interrupt after releasing the pushbutton)
	INTE 	= 0;	// Disable INT interrupts
	
	T0CS	= 0; 	// Timer0 used to count FOSC/4 pulses (1 MHz)
	PSA		= 0;	// Prescaler is used, so every 32 us Timer0 will be incremented by one from the initial preloaded value. 
	PS2		= 1;	// "100" is a prescaler of 1:32
	PS1		= 0;
	PS0		= 0;

	T0IE 	= 0;	// Timer0 interrupts not allowed when initiating the microcontroller	
					// Activating the Timer0 interrupts will be done in "Set_Value_High" and in "Set_Value_Low" states
	GIE 	= 1;	// Global interrupts allowed 

	Time_Out = 0;	

	lcd_init(FOURBIT_MODE); 	/* Initialise the LCD module  */
	lcd_clear();
	lcd_home();						// select line 1
	lcd_puts("Press Button ...");
	lcd_home2();					// select line 2 to write the result 
	lcd_puts(" .. to start betting ..");	

}

//***************************************************************************
//Interrupt Service Routine (only for the generation of the FSM's CLK sisgnal 
//                           using Timer0)
//***************************************************************************

static void interrupt service_routine (void){	//Interrupt is the key word here
	if(T0IF) {			// Is it a Timer0 overflow?
								
		T0IF = 0;   // clear interrupt flag
		T0IE = 0;	// disable interrupts when done
		Time_Out = 1; 	// Set the flag
        }
}

//*****************************************************************************
//Scan inputs 
//*****************************************************************************
char read_inputs (void){
	char value = 0;
	if (pushbutton_L == 0){
		value = 1;
 	}
	return (value);
}
	
	